import sys, os, re
from collections import Counter
from collections.abc import Iterable


# this file is in theory not necessary and should be provided by the user
# but we provide some examples here that are used in the tests
# For the tests, do not put these in a fixture, because they get set in a class level container and key

def key_type_enforcer(instance_or_type,
                      enforced_type: type,
                      enforced_key: str,
                      attrs: dict = None):
    member_object = getattr(instance_or_type, enforced_key, None)
    if member_object is None:
        if attrs is not None:
            member_object = attrs.get(enforced_key, None)
    if member_object is None:
        return False
    else:
        return issubclass(type(member_object), enforced_type)
    pass


def min_value(instance_or_type,
              enforced_key: str,
              hard_floor):
    member_object = getattr(instance_or_type, enforced_key, None)
    if member_object is None:
        return False
    return member_object > hard_floor

def min_list_type_counter(instance_or_type,
                          list_name: str,
                          min_counter: Counter,
                          attrs: dict = None):
    member_object = getattr(instance_or_type, list_name, None)
    if member_object is None:
        if attrs is not None:
            member_object = attrs.get(list_name, None)
    if member_object is None:
        return False
    else:
        if isinstance(member_object, Iterable):
            return Counter(type(x) for x in member_object) >= min_counter
        else:
            return False



import types
from typing import Type
from functools import wraps

from has_enforced_rules import HasEnforcedRules, EnforcedFunctions


def raise_error_if_false(enforced_function: types.FunctionType,
                         exception_type: Type[BaseException],
                         on_class: bool = True,
                         extra_info: str = None):
    if extra_info is None:
        extra_info = ''

    def raise_when_false(cls, function_name):
        @wraps(enforced_function)
        def wrapped_raise_when_false(*args, **kwargs):
            if on_class:
                # if on a class, we need to check if attributes dictionary got passed
                # if it did it needs to go on the kwargs
                if len(args) > 1:
                    # attrs should be the only dictionary argument
                    dict_args = [x for x in args if isinstance(x, dict)]
                    if len(dict_args) > 1:
                        raise ValueError("Class enforcing function was called with more than one dictionary argument "
                                         "(reserved for attributes)")
                    else:
                        kwargs['attrs'] = dict_args[0]
                        args = tuple(x for x in args if x != dict_args[0])
            if enforced_function(*args, **kwargs) is False:
                if on_class:
                    error_str = (f"{extra_info} {cls.__class__.__name__} fails class check {function_name} will "
                                 f"raise {exception_type.__qualname__}").strip()
                else:
                    error_str = (f"{extra_info} {cls.__class__.__name__} fails instance check {function_name} will "
                                 f"raise {exception_type.__qualname__}").strip()
                exception_instance = eval(f'{exception_type.__qualname__}("{error_str}")')
                raise exception_instance

        return wrapped_raise_when_false

    if on_class:
        def wrapped_class(cls):
            if 'HasEnforcedRules' not in type(cls).__name__:
                # not issubclass(type(cls), HasEnforcedRules):
                raise AttributeError(
                    f"{cls.__class__.__name__} must be of type {HasEnforcedRules.__class__.__name__} in order to "
                    f"use the decorator raise_error_if_false")
            function_name = str(enforced_function)
            func_to_add = raise_when_false(cls, function_name)
            EnforcedFunctions.add_enforce_function_to_class(cls.__name__, func_to_add)
            func_to_add(cls)
            return cls

        return wrapped_class
    else:
        def function_that_adds_check_instance_class(cls):
            if 'HasEnforcedRules' not in type(cls).__name__:
                # not issubclass(type(cls), HasEnforcedRules):
                raise AttributeError(
                    f"{cls.__class__.__name__} must be of type {HasEnforcedRules.__class__.__name__} in order to use "
                    f"the decorator raiseErrorIfFalse on instance creation")
            function_name = str(enforced_function)
            EnforcedFunctions.add_enforce_function_to_instance(cls.__name__,
                                                               raise_when_false(cls, function_name))
            # this now needs to be checked at every instance not on class type instantiation
            return cls

        return function_that_adds_check_instance_class
    pass


def raise_if_false_on_class(enforced_function: types.FunctionType,
                            exception_type: Type[BaseException] = Type[AttributeError],
                            extra_info: str = None):
    return raise_error_if_false(enforced_function,
                                exception_type,
                                on_class=True,
                                extra_info=extra_info)


def raise_if_false_on_instance(enforced_function: types.FunctionType,
                               exception_type: Type[BaseException] = Type[ValueError],
                               extra_info: str = None):
    return raise_error_if_false(enforced_function,
                                exception_type,
                                on_class=False,
                                extra_info=extra_info)



import pytest
import types
import sys
from functools import partial
from collections import Counter
from collections.abc import Iterable
# make sure to add to python path!
from has_enforced_rules import HasEnforcedRules
from decorators import raise_if_false_on_class, raise_if_false_on_instance
from predicates import key_type_enforcer, min_value, min_list_type_counter

def test_class_method_ok_1():
    check_method_present = partial(key_type_enforcer,
                                   enforced_type=types.FunctionType,
                                   enforced_key='library_functionality')
    try:
        @raise_if_false_on_class(check_method_present, AttributeError)
        class HasCorrectMethodClass(metaclass=HasEnforcedRules):
            def library_functionality(self):
                return 1
    except AttributeError as ex:
        pytest.fail(f"Failed {sys._getframe().f_code.co_name} with AttributeError {str(ex)}")


def test_class_method_fails_1():
    # with pytest.raises(AttributeError):
    check_method_present = partial(key_type_enforcer,
                                   enforced_type=types.FunctionType,
                                   enforced_key='library_functionality')
    try:
        @raise_if_false_on_class(check_method_present, AttributeError)
        class MissingCorrectMethodClass(metaclass=HasEnforcedRules):
            pass
    except AttributeError as ex:
        pass
    else:
        pytest.fail(f"Failed {sys._getframe().f_code.co_name} should have raised AttributeError but did not")

def test_class_attribute_ok_2():
    geq_type_dict = {str: 1, int: 2, float: 1}
    static_member_name = "STATIC_SET"
    check_has_enough_of_type = partial( min_list_type_counter,
                                        list_name=static_member_name,
                                        min_counter=Counter(geq_type_dict))
    try:
        @raise_if_false_on_class(check_has_enough_of_type,
                                 AttributeError,
                                 f"f{static_member_name} must have these or more {geq_type_dict}")
        class HasSufficientStaticSetClass(metaclass=HasEnforcedRules):
            STATIC_SET = ("Test", 10, 40, 50, 45.5, 60.0, '3', 'i', BaseException())
            pass
    except AttributeError as ex:
        pytest.fail(f"Failed {sys._getframe().f_code.co_name} with AttributeError {str(ex)}")

def test_class_method_fails_2():
    geq_type_dict = {str: 1, int: 2, float: 1}
    with pytest.raises(AttributeError):
        @raise_if_false_on_class(partial(min_list_type_counter,
                                         list_name='STATIC_LIST',
                                         min_counter=Counter({str: 1, int: 2, float: 1})),
                                 AttributeError,
                                 f"STATIC_LIST member must have these or more {geq_type_dict}")
        class HasInsufficientStaticSetClass(metaclass=HasEnforcedRules):
            STATIC_LIST = ("Test", 10, 40, 50, '3', 'i', BaseException())
            pass


def test_instance_method_ok_1():
    try:
        @raise_if_false_on_instance(partial(key_type_enforcer,
                                            enforced_type=int,
                                            enforced_key='x'),
                                    AttributeError)
        class HasCorrectInstanceMemberVariable(metaclass=HasEnforcedRules):
            def __init__(self, value=20):
                self.x = value

        a = HasCorrectInstanceMemberVariable()
    except AttributeError as ex:
        pytest.fail(f"Failed {sys._getframe().f_code.co_name} with AttributeError {str(ex)}")
    except BaseException as ex:
        pytest.fail(f"Failed {sys._getframe().f_code.co_name} with Exception {type(ex)} {str(ex)}")


def test_instance_method_fails_1():
    with pytest.raises(AttributeError):
        @raise_if_false_on_instance(partial(key_type_enforcer,
                                            enforced_type=int,
                                            enforced_key='x'),
                                    AttributeError)
        class HasIncorrectInstanceMemberVariable(metaclass=HasEnforcedRules):
            def __init__(self, value=20):
                self.y = value

        a = HasIncorrectInstanceMemberVariable()


def test_instance_method_fails_2():
    with pytest.raises(AttributeError):
        @raise_if_false_on_instance(partial(key_type_enforcer, enforced_type=int, enforced_key='x'), AttributeError)
        class HasDeletedInstanceMemberVariable(metaclass=HasEnforcedRules):
            def __init__(self, value=20):
                self.x = value
                del self.x

        a = HasDeletedInstanceMemberVariable()



from functools import wraps

def false_on_raise_else_true(func):
    # will be used when we 'transfer' enforced rules
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except Exception as ex:
            return False

    return func_wrapper



import types
from collections import defaultdict
import types
from collections import defaultdict
from utils import false_on_raise_else_true


class HasEnforcedRules(type):

    # def __new__(cls,
    #             name_of_class: str,
    #             baseclasses: tuple,
    #             attrs: dict
    #             ):
    #     # We allow the derived classes to create the type for the class
    #     # however they see fit and check any class level on creation of the type
    #     # hence attributes need to be passed.
    #     # Base classes will have their own call to this __new__ with their own specific attributes
    #     # provided, they too are of type HasEnforcedRules.
    #     cls_instance = super().__new__(cls, name_of_class, baseclasses, attrs)
    #     EnforcedFunctions.run_functions_applied_to_class(cls_instance, attrs)
    #     return cls_instance

    def __call__(cls,
                 *args,
                 **kwargs):
        # Create an object instance
        instance = super().__call__(*args, **kwargs)
        # We allow the derived classes to create an class instance
        # however they see fit and check any instance level checks here
        # all of them need to be checked at every instance creation!
        EnforcedFunctions.run_functions_applied_to_instance(instance)
        # print(f"Instance of {cls.__name__} checked")
        return instance


class EnforcedFunctions:
    _functions_applied_to_instance = defaultdict(set)
    _functions_applied_to_class = defaultdict(set)

    @classmethod
    def _apply_functions_applied_to_class(cls, cls_instance: type, attrs: dict = None):
        if cls._functions_applied_to_class[cls_instance.__name__]:
            for func in cls._functions_applied_to_class[cls_instance.__name__]:
                func(cls_instance, attrs)
                # print(f"PASSED {func.__name__} applied to class {cls_instance.__name__}")

    @classmethod
    def _apply_functions_applied_to_instance(cls, instance, cls_key: str):
        if cls._functions_applied_to_instance[cls_key]:
            for func in cls._functions_applied_to_instance[cls_key]:
                func(instance)
                # print(f"PASSED {func.__name__} applied to instance {str(instance)}")

    @classmethod
    def add_enforce_function_to_class(cls,
                                      cls_key: str,
                                      func):
        cls._functions_applied_to_class[cls_key].add(func)

    @classmethod
    def add_enforce_function_to_instance(cls,
                                         cls_key: str,
                                         func):
        cls._functions_applied_to_instance[cls_key].add(func)
        # print((cls_key, len(cls._functions_applied_to_instance[cls_key])))

    @classmethod
    def run_functions_applied_to_class(cls,
                                       cls_instance: type,
                                       attrs: dict = None):
        if len(cls._functions_applied_to_class) > 0:
            cls._apply_functions_applied_to_class(cls_instance, attrs)
            # the bases will get called individually with their own attrs if they are of type HasEnforcedRules
            # print(self._functions_applied_to_class)
            # classes_to_check = [cls_instance]
            # if bases is not None:
            #     classes_to_check.extend([x for x in bases if isinstance(x, type)])
            # for x in classes_to_check:
            #     cls._apply_functions_applied_to_class(x)

    @classmethod
    def run_functions_applied_to_instance(cls, instance):
        if not issubclass(type(type(instance)), HasEnforcedRules):
            raise ValueError(
                f"Attempt to check functions_applied_to_instance applied on an instance of {type(instance)}, "
                f"which is not of HasEnforcedRules type")
        # for the instance functions we do need to loop through all the bases
        if len(cls._functions_applied_to_instance) > 0:
            # print(self._functions_applied_to_instance)
            cls_keys = [type(instance).__name__]
            bases = [x.__name__ for x in type(instance).__bases__]
            if bases is not None:
                cls_keys.extend(bases)
            for cls_key in cls_keys:
                cls._apply_functions_applied_to_instance(instance, cls_key)

    @classmethod
    def get_functions_applied_instance(cls, class_name: str):
        return cls._functions_applied_to_instance[class_name]

    @classmethod
    def get_functions_applied_class(cls, class_name: str):
        return cls._functions_applied_to_class[class_name]

    @classmethod
    def revert_to_boolean_returns(cls, class_names=None):
        """
        Returns a tuple which for the supplied class names (defaults to all class names) with the first element:
        a dictionary with the class names as keys and as values the list of functions checking class structure
        but returning True if the check passed and False if the original check function had thrown an
        exception with the second element: a dictionary with the class names as keys and as values the list of
        functions checking a class instance but returning True if the check passed and False if the original check
        function had thrown an exception This is not used yet, but could be part of run-time transfer of
        rules between classes.

        :param class_names: The class names that are the keys of the original dictionary. Note that if an entire
        class hierarchy is required all base classes (that could have enforced rules) needs to be supplied :type
        class_names: a container type that allows the use of 'in', i.e. overloads __contains__
        """
        if class_names is None:
            class_names = list(cls._functions_applied_to_class.keys()) + list(
                cls._functions_applied_to_instance.keys())

        return (
            {key: [false_on_raise_else_true(func) for func in cls._functions_applied_to_class[key]] for key, _ in
             cls._functions_applied_to_class.items() if key in class_names},
            {key: [false_on_raise_else_true(func) for func in cls._functions_applied_to_instance[key]] for key, _ in
             cls._functions_applied_to_instance.items() if key in class_names}
        )
